
/*  assemble a text file according to map file A80.DAT (or as specified
 *  in the command line), outputting hexadecimal digits
 *
 *      a80 [-s] [-d] [-b] [-f name] [-q] [-?]
 */


/* set flag DEBUG to make redirection to "debug" work nicely */
/* set flag BIG_ENDIAN to make the target big-endian */

/*******************************************************************
 **                   Constant Declarations                       **
 *******************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define DATA_NAME "default.dat"             /* default data file name */

#define MAXPRGM 60000L                       /* length of bytecode array */
#define MAXPATTERN 30                 /* length of OPCODE lhs (mnemonic) */
#define MAXCODE 15                  /* length of OPCODE rhs (hex digits) */
#define MAXNAME 25                         /* length of a label (@xx...) */
#define MAXLINE 80                            /* length of an input line */

/* This is a 16-bit assembler only! */

#define HIBYTE(x) (((x) & 0xFF00) >> 8)
#define LOBYTE(x) ((x) & 0xFF)

enum {L_OFFSET, L_IMMEDIATE, L_OFFW};

long int PrgmLength;

int Spaces_On = 1;                            /* print spaces in output? */
int Quiet = 0;                         /* ignore undefined label errors? */
int Debug = 0;                              /* easy redirection to DEBUG */
int Big_Endian = 0;                    /* Big-endian word representation */
char Comment_Delimiter[5] = "//";  /* Comments always end with a newline */


struct                  /* Some arguments recognized by asm80 */
  {
    const char *text;
    int *flag;
    int action;
  } argtable[] =
{
        {       "-s",                   &Spaces_On,     2       },
        {       "--spaces-on",          &Spaces_On,     1       },
        {       "--spaces-off",         &Spaces_On,     0       },
        {       "-d",                   &Debug,         2       },
        {       "--debug",              &Debug,         1       },
        {       "--no-debug",           &Debug,         0       },
        {       "-q",                   &Quiet,         2       },
        {       "--quiet",              &Quiet,         1       },
        {       "--no-quiet",           &Quiet,         (int) 0 },
        {       "-b",                   &Big_Endian,    2       },
        {       "--big-endian",         &Big_Endian,    1       },
        {       "--little-endian",      &Big_Endian,    0       },
        {       "--no-big-endian",      &Big_Endian,    0       },
};


int START = 0x0100;              /* what physical address to assemble to */

struct OPCODE;
struct OPCODE *O_BUILD;                                /* pseudo-opcodes */
struct OPCODE *O_ORG;
struct OPCODE *O_PAD;
struct OPCODE *O_PAD_NOP;

unsigned char bytecode[MAXPRGM];            /* contains assembler output */

/*******************************************************************
 **                   Function Prototypes                         **
 *******************************************************************/

int is_name(int c);
int atoh(char s[]);                      /* convert ascii to hexadecimal */
char *find_number(char s[], int p);
                                    /* find the pth number or label in s */

typedef struct LabelNode *LabelPtr;
void newlabel(char s[], long int loc);
LabelPtr findlabel(char s[]);
void deletelabels(void);

typedef struct GapNode *GapPtr;
void newgap(char s[], long int pc, int type);
void fillgaps(unsigned char map[]);
void deletegaps(void);

long atoinc(char text[], int *index);
                                    /* find number in text, update index */
int match_codes(char pat[], char text[]);
struct OPCODE *get_opcode(char text[]);
struct OPCODE *add_opcode(char pat[], char code[]);
void deletecodes(void);
int no_codes(void);             /* are there any codes in the table yet? */
void initcodes(char *fname);              /* read codes from a text file */

char *getline(FILE *fp);
void stripws(char s[]);
long translate_one(struct OPCODE *match, long pc, unsigned char map[], char
    text[]);                                       /* return value of pc */
int translate(long pc, unsigned char map[], char text[]);
                                                  /* return change in pc */

void convertfile(FILE *infile, unsigned char bytecode[], long spc);
void outputcode(unsigned char bytecode[]);


int Atoi(char *s);                               /* decimal or hex atoi */
void *Malloc(size_t t);                          /* error-checked malloc */
void PrintLine(char *s);                         /* debug; prints line s */
void Abort(char *s);                     /* abnormal exit; delete labels */
void deletestuff(void);                           /* delete linked lists */


/*******************************************************************
 **                      Main Function                            **
 *******************************************************************/


int main(int argc, char *argv[])
  {
    int i;
    int argtry;

    for (i=1; i < argc; i++)
      {
        for (argtry = 0; argtable[argtry].text; argtry++)
          {
            if (strcmp(argv[i], argtable[argtry].text) == 0)
              {
                if (argtable[argtry].flag == 0)
                  continue;
                if (argtable[argtry].action == 0)
                  *(argtable[argtry].flag) = 0;
                else if (argtable[argtry].action == 1)
                  *(argtable[argtry].flag) = 1;
                else if (argtable[argtry].action == 2)
                  *(argtable[argtry].flag) = ! *(argtable[argtry].flag);
                else
                  {
                    fprintf(stderr, "Unrecognized action: done\n");
                    exit(9);
                  }
              }
          }

        if (argtable[argtry].text != 0) continue;

        if (strcmp(argv[i], "-f") == 0)
          initcodes(argv[++i]);
        else if (!strcmp(argv[i], "-C") || !strcmp(argv[i], "--comment"))
          {
            strncpy(Comment_Delimiter, argv[++i], sizeof(Comment_Delimiter));
            Comment_Delimiter[sizeof(Comment_Delimiter)-1] = '\0';
          }
        else if (!strcmp(argv[i], "-?") || !strcmp(argv[i], "--help"))
          {
            printf("General Table-driven Assembler\n");
            printf("  -s: add spaces to hex output (currently %s)\n",
                (Spaces_On)? "on" : "off");
            printf("  -q: suppress error diagnostics (currently %s)\n",
                (Quiet)? "on" : "off");
            printf("  -b: big-endianness of processor (currently %s)\n",
                (Big_Endian)? "on" : "off");
            printf("  -d: easily debug-piped output (currently %s)\n",
                (Debug)? "on" : "off");
            printf("  -f: next argument is additional data-file name\n");
            printf("      Current default data-file is \"%s\"\n", DATA_NAME);
            printf("  -C: next argument is comment delimiter (max %d chars)\n",
                sizeof(Comment_Delimiter));
            printf("      Current comment delimiter is \"%s\"\n",
                Comment_Delimiter);
            printf("  -v: show version information\n");
            printf("  -?: show help screen\n");
            exit(0);
          }
        else if (!strcmp(argv[i], "-v") || !strcmp(argv[i], "--version"))
          {
            printf("General Table-driven Assembler\n");
            printf("  Current version information:\n");
#ifdef __TURBOC__
            printf("  Compiled with Turbo C version %X\n", __TURBOC__);
#elif defined(__GNUC__)
            printf("  Compiled with Gnu C version %X\n", __GNUC__);
#else
            printf("  Compiled with unknown compiler\n");
#endif
            printf("  on %s\n", __DATE__);

            printf("  with the %s model\n",
#if defined(__TINY__) || defined(__SMALL__)
  "Tiny/Small"
#elif defined(__MEDIUM__)
  "Medium"
#elif defined(__COMPACT__)
  "Compact"
#elif defined(__LARGE__)
  "Large"
#elif defined(__HUGE__)
  "Huge"
#else
  "unknown"
#endif
               );
            exit(0);
          }
        else
          {
            fprintf(stderr, "Unrecognized argument: %s\n", argv[i]+1);
            fprintf(stderr, "\"asm80 -?\" for help\n");
            exit(1);
          }
      }

    if (no_codes())                  /* we haven't loaded any tables yet */
      initcodes(DATA_NAME);

    convertfile(stdin, bytecode, 0);
    fillgaps(bytecode);
    outputcode(bytecode);
    deletestuff();
    return 0;
  }

/*******************************************************************
 **                      Useful Functions                         **
 *******************************************************************/

int is_name(int c)
  {
    if (c <= ' ') return 0;
    if (c == '[' || c == ']') return 0;
    if (c == ',' || c == ':') return 0;
    if (c == '+' || c == '-') return 0;
    if (c == '(' || c == ')') return 0;
    return 1;
  }

int atoh(char s[])                       /* convert ascii to hexadecimal */
  {
  int result = 0;
  int i = 0;
    while (s[i])
      {
        result <<= 4;
        if ('0' <= s[i] && s[i] <= '9')
          result += s[i] - '0';
        else
          result += (s[i] & ~040) - 'A' + 10;
        i++;
      }
    return result;
  }

char *find_number(char s[], int p)  /* find the pth number or label in s */
  {
  static char t[MAXNAME];
  int i, j;
    for (i = 0; p; p--)
      {
        for (; s[i] && !strchr("0123456789-@$", s[i]); ++i)
          ; /* find beginning of number or end of string */

        if (!s[i])
          return NULL;

        j = 0;
        if (s[i] != '@')                               /* must be number */
          {
            do {
              t[j++] = s[i++];                     /* copy number into t */
            } while (strchr("0123456789ABCDEF", s[i]));
            t[j] = '\0';
          }
        else                                               /* is a label */
          {
            while (s[i] && is_name(s[i]))
              t[j++] = s[i++];                      /* copy label into t */
            t[j] = '\0';
          }
      }
    return t;                                               /* return it */
  }


/*******************************************************************
 **                 Label Structure Functions                     **
 *******************************************************************/

struct LabelNode
  {
    struct LabelNode *next;
    char name[MAXNAME];
    long loc;                             /* abstract location in memory */
  };

/* typedef struct LabelNode *LabelPtr; */

LabelPtr firstlabel = NULL;

void newlabel(char s[], long loc)
  {
  LabelPtr newspot;

    if (findlabel(s))
      {
        Abort("Redefinition of label");
      }
    newspot = (LabelPtr) Malloc(sizeof (struct LabelNode));
    newspot->next = firstlabel;
    strncpy(newspot->name, s, MAXNAME);
    newspot->loc = loc;
    firstlabel = newspot;
  }

LabelPtr findlabel(char s[])
  {
  LabelPtr cur;

    for (cur = firstlabel; cur; cur = cur->next)
      if (!strcmp(cur->name, s))
        return cur;
    return NULL;
  }

void deletelabels()
  {
  LabelPtr cur;
    while (firstlabel)
      {
        cur = firstlabel;
        firstlabel = firstlabel->next;
        free(cur);
      }
  }


/*******************************************************************
 **                  Gap Structure Functions                      **
 *******************************************************************/

struct GapNode
  {
    struct GapNode *next;
    char name[MAXNAME];
    long loc;                                /* physical location in map */
    int type;                           /* L_OFFSET, L_IMMEDIATE, L_OFFW */
  };

/* typedef struct GapNode *GapPtr; */

GapPtr firstgap = NULL;

void newgap(char s[], long pc, int type)
  {
  GapPtr newspot;

    newspot = (GapPtr) Malloc(sizeof (struct GapNode));
    newspot->next = firstgap;
    strncpy(newspot->name, s, MAXNAME);
    newspot->loc = pc;
    newspot->type = type;
    firstgap = newspot;
  }

void fillgaps(unsigned char map[])
  {
  GapPtr cur;
  LabelPtr temp;

    for (cur = firstgap; cur; cur = cur->next)
      {
        temp = findlabel(cur->name);
        if (!temp)
          {
            if (Quiet)                        /* insert nonsense address */
              temp = firstlabel;
            else
              {
                PrintLine(cur->name);
                Abort("No match for label");
              }
          }
        if (cur->type == L_OFFSET)
          {
            if (abs(temp->loc - (1+cur->loc+START)) > 127)
              {
                fprintf(stderr, "|%s| %li bytes\n", cur->name,
                  (temp->loc - (1+cur->loc+START)) );
                Abort("Short jump > 128 bytes");
              }
            map[cur->loc] = (temp->loc - (1+cur->loc+START)) & 0xFF;
          }
        else if (cur->type == L_IMMEDIATE)
          {
            if (Big_Endian)
              {
                map[cur->loc] = HIBYTE(temp->loc);
                map[1+cur->loc] = LOBYTE(temp->loc);
              }
            else
              {
                map[cur->loc] = LOBYTE(temp->loc);
                map[1+cur->loc] = HIBYTE(temp->loc);
              }
          }
        else if (cur->type == L_OFFW)
          {
            if (Big_Endian)
              {
                map[cur->loc] = HIBYTE(temp->loc - (cur->loc+START+2));
                map[1+cur->loc] = LOBYTE(temp->loc - (cur->loc+START+2));
              }
            else
              {
                map[cur->loc] = LOBYTE(temp->loc - (cur->loc+START+2));
                map[1+cur->loc] = HIBYTE(temp->loc - (cur->loc+START+2));
              }
          }
      }
  }

void deletegaps(void)
  {
  GapPtr temp;
    while (firstgap)
      {
        temp = firstgap;
        firstgap = firstgap->next;
        free(temp);
      }
  }


/*******************************************************************
 **               Opcode Structure Functions                      **
 *******************************************************************/

/*
 *  Opcode structures contain:
 *
 *  pattern, a string containing the assembler opcode
 *  opcode, a string containing the hexadecimal code
 *
 *  Pattern is coded as follows:
 *
 *  AZ09()   as written
 *  #        a byte (e.g., immediate)
 *  $        a word (e.g., jump or immediate)
 *
 *  Opcode is coded in half-bytes as follows:
 *
 *  0..F     as written
 *  #        byte
 *  $        word (little end first)
 *  ^        one-byte offset
 *  %        two-byte offset (little end first)
 *
 */


struct OPCODE
  {
    struct OPCODE *next;
    char pattern[MAXPATTERN];
    char code[MAXCODE];
  };

struct OPCODE *firstcode = NULL;

long atoinc(char text[], int *index)
  {                               /* find number in text, updating index */
    int neg = 0;
    long num = 0;

    if (text[*index] == '-')
      ++(*index), neg = 1;

    if (text[*index] == '@')     /* a label; legal but too big to return */
      {
        while (text[*index] && is_name(text[*index]))
          ++(*index);
        return 0x0FFF;
      }

    if (text[*index] == '$')
      {
        ++(*index);
        while (text[*index] && strchr("0123456789ABCDEF", text[*index]))
          {
            num *= 16;
            if (text[*index] <= '9')
              num += text[*index] - '0';
            else
              num += text[*index] - 'A' + 10;
            ++(*index);
          }
      }
    else
      {
        while ('0' <= text[*index] && text[*index] <= '9')
          {
            num *= 10;
            num += text[*index] - '0';
            ++(*index);
          }
      }
    if (neg)
      num = -num;
    return num;
  }

int match_codes(char pat[], char text[])
  {
  int i, j;                           /* see if text matches pattern pat */
  int opsize;                 /* length of parameters -- lower is better */

    i = j = 0;
    opsize = 0;
    while (pat[i] || text[j])
      {
        if (!(pat[i] && text[j]))             /* have run out of pattern */
          return 0;
        switch (pat[i])
          {
            case '#': { int temp = j;                            /* byte */
                        int ans = (int) atoinc(text, &j);
                        if (j == temp) return 0;
                        if (ans > 127) return 0;
                        if (ans < -128) return 0;
                        opsize += 1;
                        break;
                      }
            case '$':                                            /* word */
                      atoinc(text, &j);
                      opsize += 2;
                      break;
            default:                                     /* plain letter */
                      if (pat[i] != text[j]) return 0;
                      j++;
                      break;
          }
        i++;
      }
    return opsize + 1;                                     /* they match */
  }

struct OPCODE *get_opcode(char text[])
  {
  struct OPCODE *cur;
  int size;
  struct OPCODE *maybe = 0;
  int maybe_size;

    for (cur = firstcode; cur; cur = cur->next)
      {
        if ((size = match_codes(cur->pattern, text)) == 0)
          continue;
        if (size == 1) return cur;         /* return the smallest opcode */
        if (!maybe) { maybe = cur; maybe_size = size; }
        else if (size < maybe_size) return cur;
        else return maybe;
      }
    if (maybe) return maybe;
    return NULL;                /* no match */
  }


struct OPCODE *add_opcode(char pat[], char code[])
  {
  struct OPCODE *newspot;

    newspot = (struct OPCODE *) Malloc(sizeof (struct OPCODE));
    newspot->next = firstcode;
    strncpy(newspot->pattern, pat, MAXPATTERN);
    strncpy(newspot->code, code, MAXCODE);
    firstcode = newspot;
    return firstcode;
  }

int no_codes()
  {
    return (firstcode == NULL);
  }

void deletecodes()
  {
  struct OPCODE *cur;
    while (firstcode)
      {
        cur = firstcode;
        firstcode = firstcode->next;
        free(cur);
      }
  }

void initcodes(char *fname)
  {
  char field1[MAXPATTERN];
  char field2[MAXCODE];
  FILE *fp = fopen(fname, "rb");

    if (!fp)
      Abort("Cannot locate data file");

    if (no_codes())                      /* so these only get added once */
      {
        O_ORG = add_opcode("ORG $",".");
        O_PAD = add_opcode("PAD $",".");
        O_PAD_NOP = add_opcode("PAD $, NOP",".");
        O_BUILD = add_opcode("BUILD $",".");
      }

    while (!feof(fp))
      {
      int i, c;

        c = fgetc(fp);
        if (c == '*')                            /* comment in map file */
          {
            while ((c=fgetc(fp)) != '\n')              /* skip over line */
              if (c==EOF) break;          /* check for EOF, just in case */
            continue;                                  /* go back to top */
          }

        for (i=0; (c != EOF) && (c != '|'); i++)
          {
            field1[i] = c;
            c = fgetc(fp);
          }
        while (field1[--i] == ' ');
        field1[i+1] = '\0';
        c = fgetc(fp);

        for (i=0; (c != EOF) && (c != '\n'); i++)
          {
            field2[i] = c;
            c = fgetc(fp);
          }
        while (field2[--i] == ' ');
        field2[i+1] = '\0';

        add_opcode(field1, field2);
      }                                /* end of "while(!feof(fp))" loop */

    fclose(fp);
  }

/*******************************************************************
 **                  Translator Functions                         **
 *******************************************************************/


char *getline(FILE *fp)
  {
  static char s[MAXLINE];
  int i = -1;
  char *p;

    do
      {
        i++;
        if (i == sizeof(s))
          {
            s[i-1] = '\0';
            PrintLine(s);
            Abort("Length of input line too long");
          }
        s[i] = fgetc(fp);
      }
    while ((s[i] != '\n') && !feof(fp));
    s[i] = '\0';

    p = strstr(s, Comment_Delimiter);
    if (p != NULL)                /* remove end-of-line comments, if any */
      {
        *p = '\0';
      }

    return s;
  }

void stripws(char *s)
  {
  char t[MAXLINE];
  int i, j;

    i = j = 0;
    while (s[i])
      {
        while (s[i] == ' ')
          i++;
        while (s[i] && (s[i] != ' '))
          t[j++] = s[i++];
        t[j++] = s[i++];
      }
    t[j] = '\0';
    if (t[--j] == ' ')
      t[j] = '\0';
    strcpy(s, t);
  }


long translate_one(struct OPCODE *match, long pc, unsigned char map[], char
    text[])
  {
    int i = 0;
    int curplace = 1;
    char temp[MAXNAME];

    for (i = 0; (match->code[i] != 0) && (i < MAXCODE); i++)
      {
        switch (match->code[i])
          {
             case '#':
               strncpy(temp, find_number(text, curplace), MAXNAME);
               if (temp[0] == '@')
                 {
                   PrintLine(text);
                   Abort("Cannot convert label to byte");
                 }
               curplace++;
               map[pc++] = LOBYTE(Atoi(temp));
               break;
             case '$':
               strncpy(temp, find_number(text, curplace), MAXNAME);
               curplace++;
               if (temp[0] != '@')
                 {
                   if (Big_Endian)
                     {
                       map[pc++] = HIBYTE(Atoi(temp));
                       map[pc++] = LOBYTE(Atoi(temp));
                     }
                   else
                     {
                       map[pc++] = LOBYTE(Atoi(temp));
                       map[pc++] = HIBYTE(Atoi(temp));
                     }
                 }
               else
                 {
                   newgap(temp, pc, L_IMMEDIATE);
                   pc+=2;
                 }
               break;
             case '^':
               strncpy(temp, find_number(text, curplace), MAXNAME);
               curplace++;
               if (temp[0] != '@')
                 {
                   if (abs(Atoi(temp) - (pc+START+1) - 255) > 127)
                     {
                       Abort("Short jump over more than 128 bytes");
                     }
                   map[pc] = Atoi(temp) - (pc+START+1);
                   pc++;
                 }
               else
                 {
                   newgap(temp, pc, L_OFFSET);
                   map[pc++] = 0xFF;
                 }
               break;
             case '%':
               strncpy(temp, find_number(text, curplace), MAXNAME);
               curplace++;
               if (temp[0] != '@')
                 {
                   pc+=2;
                   if (Big_Endian)
                     {
                       map[pc-1] = HIBYTE(Atoi(temp) - (pc+START));
                       map[pc] = LOBYTE(Atoi(temp) - (pc+START));
                     }
                   else
                     {
                       map[pc-1] = LOBYTE(Atoi(temp) - (pc+START));
                       map[pc] = HIBYTE(Atoi(temp) - (pc+START));
                     }
                 }
               else
                 {
                   newgap(temp, pc, L_OFFW);
                   map[pc++] = 0xFF;
                   map[pc++] = 0xFF;
                 }
               break;
             case '0': case '1': case '2': case '3': case '4':
             case '5': case '6': case '7': case '8': case '9':
             case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
                 {
                   char temp[3];
                   temp[0] = match->code[i];
                   temp[1] = match->code[i+1];
                   temp[2] = '\0';
                   map[pc++] = atoh(temp);
                   i++;
                 }
               break;
          }                     /* end of switch */
      }                     /* end of for */
    return pc;
  }


                                                  /* return change in pc */
int translate(long pc, unsigned char map[], char text[])
  {
    char temp[MAXNAME];
    struct OPCODE *match;

    if (!text[0]) return 0;                                /* blank line */

    stripws(text);

    if (text[0] == '@')                                         /* label */
      {
        char *pp = strchr(text, ':');
        int p;

        if (!pp)
          {
            PrintLine(text);
            Abort("Label without terminating :");
          }
        p = pp - text;
        strncpy(temp, text, p);
        temp[p] = '\0';
        newlabel(temp, pc+START);
        text += p+1;                                /* to get over the : */
      }

    while (*text == ' ')
      text++;

    if (!text[0]) return 0;                               /* end of line */

    match = get_opcode(text);
    if (!match)                                              /* no match */
      {
        PrintLine(text);
        Abort("Illegal opcode");
      }
    else if (match == O_BUILD)                   /* BUILD 1024 (d), e.g. */
      {
        return Atoi(find_number(text, 1)) - pc;
      }
    else if (match == O_ORG)                        /* ORG 512 (d), e.g. */
      {
        START = Atoi(find_number(text, 1));
        return 0;
      }
    else if (match == O_PAD)                          /* PAD 2 (d), e.g. */
      {
        int i = Atoi(find_number(text, 1));      /* pad to the alignment */
        return (i - ((pc+START) % i)) % i;
      }
    else if (match == O_PAD_NOP)                     /* PAD 2, NOP, e.g. */
      {
        int cnt = Atoi(find_number(text, 1));    /* pad to the alignment */
        int i;

        cnt = (cnt - ((pc+START) % cnt)) % cnt;
        for (i=0; i < cnt; i++)
          {
            (void) translate_one(get_opcode("NOP"), pc++, map, "NOP");
          }
        return cnt;
      }
                                                         /* found opcode */
    return (translate_one(match, pc, map, text) - pc);
  }


/*******************************************************************
 **                    Control Functions                          **
 *******************************************************************/


void convertfile(FILE *infile, unsigned char bytecode[], long spc)
  {
    long pc = spc;
    char curline[MAXLINE];

    rewind(infile);

    while (!feof(infile))
      {
        strncpy(curline, getline(infile), sizeof curline);
        stripws(curline);
        pc += translate(pc, bytecode, curline);
        if (pc >= MAXPRGM)
          Abort("Program too large for memory");
      }
    PrgmLength = pc;
  }


void outputcode(unsigned char bytecode[])
  {
    long ip = 0;
    for (ip = 0; ip < PrgmLength; ip++)
      {
        if (Spaces_On && !(ip % 16))
          printf("\n");
        printf("%02X", *(bytecode+ip));
        if (Spaces_On)
          printf(" ");
      }
    if (Spaces_On)
      printf("\n");
  }


/*******************************************************************
 **                   Background Functions                        **
 *******************************************************************/

int Atoi(char *s)
  {
    long int base = (s[0] == '$') ? 16:10;    /* prefix with $ means hex */
    int neg = 0;
    long int num = 0L;
    static char digits[] = "0123456789ABCDEF";

    if (*s == '-')
      neg = 1, ++s;
    if (*s == '$')
      ++s;
    while (*s)
      {
        num = base*num + (strchr(digits, *s)-digits);
        ++s;
      }
    if (neg) num = -num;
    if (num > 65535L || num < -32767L)
      Abort("Numeric literal too long");
    return (signed int) num;
  }

void *Malloc(size_t t)
  {
  void *p;
    p = malloc(t);
    if (!p)
      Abort("Not enough memory");
    return p;
  }

void PrintLine(char *s)
  {
    if (Debug)
      printf("a\n%s\n\nu\nq\n", s);          /* included for "debug" use */
    else
      fprintf(stderr, "Line: |%s|\n", s);
  }

void Abort(char *s)
  {
    fprintf(stderr, "Fatal error: %s\n", s);
    deletestuff();
    exit(1);
  }

void deletestuff()
  {
    deletegaps();
    deletelabels();
    deletecodes();
  }



