
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <dir.h>

#define MAXBUF 1000
#define MAXNAME 32

int Debug = 0;                      /* This is the flag to control debug */

int curlabel;
int look;

int getlook(void);                     /* get look from the input buffer */
int putlook(int);                   /* put look back on the input buffer */

void assemble(const char *, ...);                     /* print to stdout */
void abortf(const char *, ...);     /* print to stderr and abort program */

void compile_prgm(void);                   /* This is the main procedure */

void compile_levA(int);                    /* Level A is semicolon/comma */
void compile_levB(int);
void compile_levC(int);                         /* Level C is assignment */
void compile_levD(int);                /* Level D is indirect assignment */
void compile_levE(int);
void compile_levF(int);
void compile_levG(int);
void compile_levH(int);                 /* Level H is addition operators */
void compile_levI(int);
void compile_levJ(int);
void compile_levK(int);                    /* Level K is unary operators */
void compile_levL(int);                         /* Level L is & operator */
void compile_levM(int);                            /* Level M is postfix */
void compile_atom(int);

int get_char(void);       /* get a single-character sequence in a string */
void compile_str(void);
void compile_func(void);
void compile_getmem(void);

int getatoi(void);

int address_of(const char *ident);
void add_variable(const char *ident);
void add_frame(void);
void remove_frame(void);
void print_globals(void);

void record_library(const char *libname);
void show_avail_libs(void);
void dump_libs(void);

      /****************************************************************
       **                        MAIN FUNCTION                       **
       ****************************************************************/

int main(int argc, char *argv[])
  {
    int i;
    for (i=1; i < argc; i++)
      {
        if (strcmp(argv[i], "-d") == 0)
          Debug = !Debug;
        else if (strcmp(argv[i], "-?") == 0)
          {
            printf("NEW Language Compiler, C test version\n");
            printf("  -d: emit debugging information to stderr"
                   " (currently %s)\n", (Debug)? "on" : "off");
            printf("  -l: include a system library (module)\n");
            show_avail_libs();
            printf("  -?: show help screen\n");
            exit(0);
          }
        else if (strncmp(argv[i], "-l", 2) == 0)
          {
            record_library(argv[i]+2);
          }
        else
          {
            fprintf(stderr, "Unrecognized option; \"day9 -?\" for help\n");
            exit(1);
          }
      }
    if (Debug) fprintf(stderr, "Debugging is ON\n");
    compile_prgm();                           /* do the "real work" here */
    return 0;
  }


      /****************************************************************
       **                  COMPILER CONTROL FUNCTION                 **
       ****************************************************************/

void compile_prgm(void)
  {
    add_frame();
    assemble("START PROGRAM (@start)");   /* This is the standard header */
    assemble("@0:");
    assemble("PUSH  BP");             /* The whole program is a function */
    assemble("MOV   BP, SP");
    getlook();
    curlabel = 1;
    compile_levA(0);
    if (look != EOF)
      {
        fprintf(stderr, "End of program before EOF; next character: %c\n",
            look);
        fprintf(stderr, "Continuing to compile output...\n");
      }
    remove_frame();
    assemble("MOV   SP, BP");
    assemble("POP   BP");
    assemble("RET");
    assemble("@start:");

    dump_libs();                             /* Paste library code here. */

    assemble("CALL  @0");
    assemble("END PROGRAM (0)");

    assemble("@i2b:");
    assemble("OR    AX, AX");
    assemble("JZ    @i2b2");
    assemble("MOV   AX, 1");
    assemble("@i2b2:");
    assemble("RET");

    print_globals();
    assemble("@__heap:");
  }

/*
 *  The last line puts the global variable ^_heap at the end of the
 *  program.  This is a very important hook for the memory allocation
 *  functions in STDLIB.NEW.
 */


      /****************************************************************
       **                   LOOK BUFFER FUNCTIONS                    **
       ****************************************************************/

static int lookbuf[MAXBUF];
static int lookptr = 0;

int getlook(void)
  {
    while (lookptr == 0)                        /* out of buffered input */
      {
        int tempbuf[MAXBUF];
        int i = 0;
        int inquotes = 0;
        int inchotes = 0;

        for (look = getchar(); look != '\n'; look = getchar())
          {                        /* read a line into tempbuf, forwards */
            if (look == EOF)
              {
                tempbuf[i++] = look;
                break;           /* make sure the EOF goes in the buffer */
              }
            if ( !inquotes && !inchotes )
              {
                while ((look == ' ') || (look == '\t'))
                  look = getchar();
                if (look == '$')                    /* two-char operator */
                  {
                    look = getchar()+128;    /* just to simplify parsing */
                    if (look == '/'+128)
                      {                /* ...it's a comment; ignore line */
                        while (getchar() != '\n')
                          ;                          /* read and discard */
                        break;                  /* break out of FOR loop */
                      }
                  }
              }
            else if (look == '\\')                   /* inside a literal */
              {
                tempbuf[i++] = '\\';
                tempbuf[i++] = getchar();             /* escape sequence */
                continue;
              }

            if (look == '"' && !inchotes)
              inquotes = !inquotes;     /* ignore space except in quotes */
            else if (look == '\'' && !inquotes)
              inchotes = !inchotes;     /* ignore space except in quotes */

            tempbuf[i++] = look;             /* store char in the buffer */
          }

        if ((inquotes || inchotes) && (look == '\n'))
          {
            abortf("End of line encountered inside quoted literal\n");
          }

        while (i)               /* copy the line into lookbuf, backwards */
          {
            lookbuf[lookptr++] = tempbuf[--i];
          }
      }                                 /* end of "while (lookptr == 0)" */
    look = lookbuf[--lookptr];    /* get the next input-buffer character */
    return (look != EOF);                             /* and run with it */
  }

int putlook(int c)
  {
    if (lookptr == MAXBUF-1)
      abortf("Out of buffer space for lookahead... aborting\n");
    else
      lookbuf[lookptr++] = c;
    return 1;
  }


      /****************************************************************
       **                  COMPILER LEVEL FUNCTIONS                  **
       ****************************************************************/

void compile_levA(int curfunc)           /* curfunc is a pointer to "#0" */
  {
    do
      {
        while (look == ';')                    /* allow null expressions */
          getlook();
        if (look == EOF)                  /* even at the end of the file */
          break;
        compile_levB(curfunc);
      }
    while (look == ';');
  }

void compile_levB(int cf)
  {
    compile_levC(cf);
    if (look == '?')                            /* if-then-else operator */
      {
        int jmplabel;
        if (Debug) fprintf(stderr, "FOUND IF STATEMENT ?\n");
        getlook();                                  /* read over the '?' */
        assemble("OR   AX, AX");
        assemble("JNZ  @%i", curlabel);
        jmplabel = curlabel+1;
        assemble("JMP  @%i", jmplabel);
        assemble("@%i:", curlabel);
        curlabel += 2;
        compile_levB(cf);                       /* compile "if 1" branch */
        if (look == ':')                    /* "if 0" branch is optional */
          {
            getlook();                              /* read over the ':' */
            assemble("JMP  @%i", curlabel);
            assemble("@%i:", jmplabel);
            jmplabel = curlabel;
            curlabel++;
            compile_levB(cf);                   /* compile "if 0" branch */
          }
        assemble("@%i:", jmplabel);
      }
  }

void compile_levC(int cf)
  {
    char buf[MAXBUF];
    int i;
    i = 0;
    if (look == '^')
      {
        buf[i++] = look;
        getlook();
      }
    while (isalpha(look) || (look == '_'))
      {
        buf[i++] = look;
        getlook();
      }
    buf[i] = '\0';                          /* null-terminate the string */
    if ((i == 0) || (look != '='))         /* not (^)[alpha string] =... */
      {
        if ((look == ';') && address_of(buf) == 0)
          {
            add_variable(buf);
            return;
          }
        putlook(look);
        while (i)
          putlook(buf[--i]);                   /* so put the string back */
        getlook();
        compile_levD(cf);
      }
    else                           /* if we really have an assignment... */
      {
        if (address_of(buf) == 0)       /* see if buf is a variable name */
          add_variable(buf);               /* if not, add it to the list */
        getlook();                                  /* get the next char */
        compile_levC(cf);            /* compile right side of assignment */
        if (buf[0] != '^')
          assemble("MOV   WORD PTR [BP+%i], AX", address_of(buf));
        else
          assemble("MOV   WORD PTR [@_%s], AX", buf+1);
      }
  }

void compile_levD(int cf)
  {
    compile_levE(cf);
    if (look == '@')
      {
        getlook();
        if (look != '=')           /* not an assignment; return directly */
          return;
        getlook();
        assemble("PUSH  AX");
        compile_levD(cf);            /* compile right side of assignment */
        assemble("POP   BX");
        assemble("SHL   BX, 1");         /* dereferencing a user pointer */
        assemble("MOV   WORD PTR [BX], AX");
      }
  }

void compile_levE(int cf)
  {
    compile_levF(cf);
    if (look == ('|'+128))
      {
        int temp = curlabel;
        curlabel += 2;
        getlook();
        assemble("OR    AX, AX");
        assemble("JZ   @%i", temp);
        assemble("JMP   @%i", temp+1);       /* if true, skip other half */
        assemble("@%i:", temp);
        compile_levE(cf);
        assemble("@%i:", temp+1);
        assemble("OR    AX, AX");
        assemble("JZ    @%i", curlabel);              /* if true, make 1 */
        assemble("MOV   AX, 1");                    /* if false, leave 0 */
        assemble("@%i:", curlabel);
        curlabel++;
      }
  }

void compile_levF(int cf)
  {
    compile_levG(cf);
    if (look == ('&'+128))
      {
        int temp = curlabel;
        curlabel += 2;
        getlook();
        assemble("OR    AX, AX");
        assemble("JNZ   @%i", temp);
        assemble("JMP   @%i", temp+1);      /* if false, skip other half */
        assemble("@%i:", temp);
        compile_levF(cf);
        assemble("@%i:", temp+1);
        assemble("OR    AX, AX");
        assemble("JZ    @%i", curlabel);              /* if true, make 1 */
        assemble("MOV   AX, 1");                    /* if false, leave 0 */
        assemble("@%i:", curlabel);
        curlabel++;
      }
  }

void compile_levG(int cf)                        /* relational operators */
  {
    int kind;
    compile_levH(cf);
    while ((look == '='+128) || (look == '!'+128) ||
           (look == '<') ||     (look == '>')     ||
           (look == '<'+128) || (look == '>'+128))
      {
        kind = look;
        getlook();
        assemble("PUSH  AX");
        compile_levH(cf);
        assemble("POP   BX");
        assemble("CMP   BX, AX");  /* compare first to second; set flags */
        assemble("MOV   AX, 1");    /* get set up without changing flags */
        switch (kind)
          {
            case ('='+128): assemble("JE    @%i", curlabel);  break;
            case ('!'+128): assemble("JNE   @%i", curlabel);  break;
            case ('>'):     assemble("JG    @%i", curlabel);  break;
            case ('<'):     assemble("JL    @%i", curlabel);  break;
            case ('>'+128): assemble("JGE   @%i", curlabel);  break;
            case ('<'+128): assemble("JLE   @%i", curlabel);  break;
          }
        assemble("XOR   AX, AX");                   /* clear AX to zero */
        assemble("@%i:", curlabel);            /* now AX is one or zero */
        curlabel++;
      }
  }

void compile_levH(int cf)
  {
    compile_levI(cf);
    while ((look == '+') || (look == '-'))
      {
        int kind = look;
        getlook();
        assemble("PUSH  AX");
        compile_levI(cf);
        assemble("POP   BX");
        if (kind == '+')
          {
            assemble("ADD   AX, BX");
          }
        else if (kind == '-')
          {
            assemble("SUB   AX, BX");
            assemble("NEG   AX");    /* since they're in the wrong order */
          }
      }
  }

void compile_levI(int cf)
  {
    compile_levJ(cf);
    while ((look == '*') || (look == '/') || (look == '%'))
      {
        int kind = look;
        getlook();
        assemble("PUSH  AX");
        compile_levJ(cf);
        if (kind == '*')
          {
            assemble("POP   BX");
            assemble("IMUL  BX");
          }
        else if ((kind == '/') || (kind == '%'))
          {
            assemble("MOV   BX, AX");
            assemble("POP   AX");
            assemble("XOR   DX, DX");
            assemble("IDIV  BX");      /* Don't worry about the details. */
            if (kind == '%')
              assemble("MOV  AX, DX");
          }
      }
  }

void compile_levJ(int cf)         /* the arithmetic AND/OR/XOR operators */
  {
    compile_levK(cf);
    while ((look == '&') || (look == '|') || (look == '^'))
      {
        int kind = look;
        getlook();
        assemble("PUSH  AX");
        compile_levK(cf);
        assemble("POP   BX");
        if (kind == '&')
          {
            assemble("AND   AX, BX");
          }
        else if (kind == '|')
          {
            assemble("OR    AX, BX");
          }
        else if (kind == '^')
          {
            assemble("XOR   AX, BX");
          }
      }
  }

void compile_levK(int cf)
  {
    if ((look == '*') || (look == '-') || (look == '!') || (look == '?'))
      {
        int kind = look;
        int mylabel;
        if (kind == '?')
          {
            mylabel = curlabel;
            curlabel++;
            if (Debug) fprintf(stderr, "FOUND WHILE STATEMENT ?\n");
            assemble("@%i:", mylabel);
          }
        getlook();
        compile_levK(cf);
        if (kind == '*')
          {
            assemble("MOV   BX, AX");
            assemble("SHL   BX, 1");
            assemble("MOV   AX, WORD PTR [BX]");
          }
        else if (kind == '-')
          {
            assemble("NEG   AX");
          }
        else if (kind == '!')
          {
            assemble("CALL  @i2b");
            assemble("DEC   AX");          /* from 0/1 to -1/0 */
            assemble("CALL  @i2b");
          }
        else if (kind == '?')      /* This is the while-loop expression. */
          {
            int nextlabel = curlabel;
            curlabel += 2;
            if (look != ':')
              {
                abortf("Unrecognized character following while clause: %c\n",
                    look);
              }
            getlook();
            assemble("OR    AX, AX");           /* We use JNZ/JMP rather */
            assemble("JNZ   @%i", nextlabel);   /* than JZ for technical */
            assemble("JMP   @%i", nextlabel+1);              /* reasons. */
            assemble("@%i:", nextlabel);
            compile_levB(cf);                   /* Don't want levA here! */
            assemble("JMP   @%i", mylabel);
            assemble("@%i:", nextlabel+1);
          }
      }
    else
      compile_levL(cf);
  }

void compile_levL(int cf)
  {
    char buf[MAXBUF];
    int i;
    if (look != '&')
      {
        compile_levM(cf);
        return;
      }

    i = 0;
    getlook();
    if (look == '^')
      {
        buf[i++] = look;
        getlook();
      }
    while (isalpha(look) || (look == '_'))
      {
        buf[i++] = look;
        getlook();
      }
    if (i == 0)
      abortf("Expected a variable name after '&': %c\n", look);
    buf[i] = '\0';
    if (address_of(buf) == 0)         /* This is cut-and-paste from levC */
      add_variable(buf);
    if (buf[0] == '^')
      {
        assemble("MOV   AX, @_%s", buf+1);
      }
    else
      {
        assemble("LEA   BX, [BP+%i]", address_of(buf));
        assemble("MOV   AX, BX");
      }
    assemble("SHR   AX, 1");              /* Remember the shifting part! */
  }

void compile_levM(int cf)
  {
    compile_atom(cf);
    while ((look == '(') || (look == '['))
      {
        if (look == '[')
          {
            assemble("PUSH  AX");
            getlook();
            compile_levA(cf);
            if (look != ']')
              abortf("Expected closing bracket ']': %c\n", look);
            getlook();
            assemble("POP   BX");
            assemble("ADD   BX, AX");
            assemble("SHL   BX, 1");
            assemble("MOV   AX, WORD PTR [BX]");
          }
        else if (look == '(')
          {
/*
 *  The way we handle functions with arguments is really kludgy.
 *  A "real" compiler would compile the arguments in the right order to
 *  begin with, instead of switching them all around and doing lots of
 *  pushing and popping.  But that approach requires some pretty
 *  sophisticated "look-ahead" techniques, and we don't really need our
 *  code to be optimized - it just needs to work.
 */
            int i, args = 0;
            if (Debug) fprintf(stderr, "FOUND FUNCTION CALL (\n");
            assemble("PUSH  AX");          /* mov AX into a holding spot */
            getlook();
            while (look != ')')
              {
                compile_levB(cf);               /* compile the arguments */
                assemble("POP   SI");      /* keep func-ptr on stack top */
                assemble("PUSH  AX");     /* push args onto the stack... */
                assemble("PUSH  SI");           /* ...in the wrong order */
                args++;
                if (look == ',')
                  getlook();
                else if (look != ')')
                  abortf("Unexpected character in arguments: %c\n", look);
              }
            getlook();
            assemble("POP   BX");      /* get back the func-ptr (old AX) */
            if (args > 1)
              assemble("MOV   SI, SP"); /* -this is how we get the args- */
            for (i=0; i < args/2; i++)    /* -in the right order again-  */
              {
                assemble("MOV   AX, WORD PTR [SI+%i]", 2*i);
                assemble("XCHG  AX, WORD PTR [SI+%i]", 2*(args-i-1));
                assemble("MOV   WORD PTR [SI+%i], AX", 2*i);
              }
            assemble("SHL   BX, 1");       /* dereference a user pointer */
            assemble("CALL  BX");                /* to call the function */
            if (args)
              assemble("ADD   SP, %i", 2*args); /* and restore the stack */
          }
      }
  }

void compile_atom(int curfunc)
  {
    if (isdigit(look))                                         /* number */
      {
        int i = getatoi();
        assemble("MOV   AX, %i", i);
      }
    else if (isalpha(look) || (look == '_') || (look == '^'))    /* name */
      {
        char buf[MAXBUF];
        int i;
        i = 0;
        if (look == '^')       /* this means a global variable or getmem */
          {
            buf[i++] = look;
            getlook();
            if (isdigit(look))                  /* whoops, it's a getmem */
              {
                compile_getmem();
                return;                            /* so return directly */
              }
          }
        while (isalpha(look) || (look == '_'))           /* get the name */
          {
            buf[i++] = look;
            getlook();
          }
        buf[i] = '\0';
        if (address_of(buf) == 0)
          {
            add_variable(buf);
/*
 *  The user probably didn't mean for this to happen, and it's probably
 *  going to cause a crash anyway, so let's warn him.  The most common
 *  cause of this warning would be something like
 *      mybuffer = malloc(100)
 *  instead of
 *      mybuffer = ^malloc(100)
 *  which can be really hard to detect.
 */
             fprintf(stderr, "Variable %s implicitly declared in"
                 " expression\n", buf);
          }
        if (buf[0] != '^')
          assemble("MOV   AX, WORD PTR [BP+%i]", address_of(buf));
        else
          assemble("MOV   AX, WORD PTR [@_%s]", buf+1);
      }
    else if (look == '\"')                                     /* string */
      compile_str();
    else if (look == '\'')                                       /* char */
      {
        getlook();
        assemble("MOV   AX, %i", get_char());
        if (look != '\'')
          abortf("Close-quote expected after character literal\n");
        getlook();
      }
    else if (look == '(')                    /* parenthesized expression */
      {
        if (Debug) fprintf(stderr, "FOUND EXPRESSION (\n");
        getlook();
        compile_levA(curfunc);
        if (look != ')')
          abortf("Expected closing parenthesis: %c\n", look);
        getlook();
      }
    else if (look == '{')                                    /* function */
      compile_func();
    else if (look == '#')                                    /* argument */
      {
        int i;
        getlook();
        i = getatoi();
        if (i == 0)                         /* special case function ptr */
          {
            assemble("MOV   AX, @%i", curfunc);
            assemble("SHR   AX, 1");
          }
        else                                        /* numbered argument */
          assemble("MOV   AX, WORD PTR [BP+%i]", (2*i)+2);
      }
    else
      {
        abortf("Expected an atom: %c\n", look);
      }
  }

int get_char(void)
  {
    int value = 0;
    if (look == '\\')
      {
        getlook();
        if (look == '\\') value = '\\';
        else if (look == '\'') value = '\'';
        else if (look == '\"') value = '\"';
        else if (look == 'n') value = 10;
        else if (look == 't') value = 9;
        else if (look == 'b') value = 32;
        else if (look == 'z') value = -999;     /* leaves off ending NUL */
        else
          {
            int i;
            value = 0;
            for (i=0; (i<4) && isxdigit(look); i++)
              {
                if (look <= '9')
                {
                  value = value*16 + look - '0';
                }
                else
                {
                  value = value*16 + toupper(look) - 'A' + 10;
                }
                getlook();
              }
            putlook(look);
          }
      }
    else
      value = look;
    getlook();
    return value;
  }

void compile_str(void)
  {
    int ch = 0;
    assemble("JMP   @%i", curlabel+1);
    assemble("PAD 2, NOP");            /* This insures correct alignment */
    assemble("@%i:", curlabel);           /* (remember word-addressing!) */
    while (look == '"')                  /* concatenate adjacent strings */
      {
        getlook();                              /* get character after " */
        while (look != '"')
          {
            ch = get_char();
            if (ch != -999)
              assemble("DATA = %i", ch);
          }
        getlook();                      /* get character after closing " */
      }
    if (ch != -999)    /* When string ends with token '\z', don't NUL it */
      assemble("DATA = 0");
    assemble("@%i:", curlabel+1);
    assemble("MOV   AX, @%i", curlabel);
    assemble("SHR   AX, 1");
    curlabel += 2;
  }

void compile_func(void)
  {
    int funclabel;
    funclabel = curlabel;
    curlabel += 2;
    assemble("JMP   @%i", funclabel+1);        /* Jump around this block */
    assemble("PAD 2, NOP");
    assemble("@%i:", funclabel);
    assemble("PUSH  BP");                /* here's that function header! */
    assemble("MOV   BP, SP");
    add_frame();                       /* make a new "frame" (namespace) */
    getlook();                                                  /* get { */
    compile_levA(funclabel);             /* Level A is the function body */
    if (look != '}')
      abortf("Expected '}' following function: %c\n", look);
    getlook();                                                  /* get } */
    remove_frame();                              /* clean up the "frame" */
    assemble("MOV   SP, BP");
    assemble("POP   BP");
    assemble("RET");                    /* ...and the function ends here */
    assemble("@%i:", funclabel+1);                    /* ...jump to here */
    assemble("MOV   AX, @%i", funclabel);
    assemble("SHR   AX, 1");              /* Here's the function pointer */
  }

void compile_getmem(void)
  {
    char buf[MAXNAME];
    int i = 0;
    buf[i++] = '^';
    while (isdigit(look))                              /* get the number */
      {
        buf[i++] = look;
        getlook();
      }
    buf[i] = '\0';
    add_variable(buf);              /* this remembers to alloc the space */
    assemble("MOV   AX, @%i", curlabel);
    assemble("SHR   AX, 1");              /* because it's a user pointer */
    curlabel++;
  }


int getatoi(void)
  {
    char buf[MAXBUF];
    int i = 0;
    while (isdigit(look))
      {
        buf[i++] = look;
        getlook();
      }
    buf[i] = '\0';
    return atoi(buf);
  }


      /****************************************************************
       **                   VARIABLE NAME DATABASE                   **
       ****************************************************************/

struct VarType
  {
    char name[MAXNAME];
    int address;
    struct VarType *next;
  };

struct VarType *VarTree = NULL;

int address_of(const char *ident)
  {
    struct VarType *cur;
    for (cur = VarTree; cur != NULL; cur = cur->next)
      {
        if (ident[0] != '^')                    /* it's a local variable */
          if (strcmp(cur->name, "@FRAME") == 0)
            break;                             /* only search this frame */
        if (strcmp(cur->name, ident) == 0)    /* found a match for ident */
          return cur->address;
      }                                  /* only go up to the last frame */
    return 0;
  }

void add_variable(const char *ident)
  {
    struct VarType *cur;
    cur = VarTree;
    VarTree = (struct VarType *) malloc(sizeof (struct VarType));
    VarTree->next = cur;
    strncpy(VarTree->name, ident, MAXNAME);
    VarTree->name[MAXNAME-1] = '\0';
    if (ident[0] != '^')                        /* it's a local variable */
      {
        VarTree->address = cur->address-2;
        assemble("SUB   SP, 2");
      }

 /*   That last statement requires some explanation.  Our variables go on
  *   the stack, so we have to lower the stack top accordingly.  But we
  *   are also using the stack for intermediate results, so if we got an
  *   expression like
  *                       1+(x=2)
  *                                  , where x is being implicitly declared
  *   for the first time, our compiler as it stands now will push 1, lower
  *   SP, put 2 where 1 had been, and pop garbage.  (Work it out on paper
  *   if it's not clear yet.)  In short, declaring variables in the middle
  *   of expressions will cause the compiler to produce buggy code.
  *     There is no simple solution to this problem on the compiler end.
  *   It can be avoided by initializing variables with simple expressions
  *   before using them.
  */

    else                                           /* it begins with '^' */
      {
        if (isdigit(ident[1]))                          /* it's a getmem */
          VarTree->address = curlabel;            /* sneaky, but workful */
        else                                   /* it's a global variable */
          VarTree->address = 1;
        for (cur = VarTree; cur->next != NULL; cur = cur->next)
          ;
        cur->next = VarTree;         /* rotate first node to end of list */
        cur = VarTree->next;               /* behind the top-level frame */
        VarTree->next = NULL;                                /* YANETUTY */
        VarTree = cur;
      }
  }                                             /* end of add_variable() */

void add_frame(void)
  {
    struct VarType *cur;
    cur = VarTree;
    VarTree = (struct VarType *) malloc(sizeof (struct VarType));
    VarTree->next = cur;
    strcpy(VarTree->name, "@FRAME");
    VarTree->address = 0;
  }

void remove_frame(void)         /* remove all variables up to next frame */
  {
    struct VarType *cur;
    cur = VarTree;
    while (strcmp(cur->name, "@FRAME") != 0)
      {
        VarTree = cur->next;
        free(cur);
        cur = VarTree;
      }
    VarTree = cur->next;
    free(cur);
  }

void print_globals(void)       /* assemble DATA instructions for globals */
  {
    struct VarType *cur;

    assemble("PAD 2, NOP");

    cur = VarTree;
    while (cur != NULL)
      {
        if (isdigit((cur->name)[1]))             /* it's really a getmem */
          {
            int i;
            assemble("@%i:", cur->address);
            for (i=0; i<atoi(cur->name+1); i++)
              assemble("DATA");                 /* assemble lots of DATA */
          }
        else
          {                                      /* ^_heap is predefined */
            if (strcmp(cur->name, "^_heap") != 0)
              {
                assemble("@_%s:", (cur->name)+1);   /* leave off the '^' */
                assemble("DATA");
              }
          }
        VarTree = cur->next;
        free(cur);
        cur = VarTree;
      }
  }


      /****************************************************************
       **                      OUTPUT FUNCTIONS                      **
       ****************************************************************/

void abortf(const char *format, ...)
  {
    va_list v;
    va_start(v, format);
    vfprintf(stderr, format, v);
    va_end(v);
    exit(1);
  }

void assemble(const char *format, ...)
  {
    va_list v;
    va_start(v, format);
    if (format[0] != '@')
      {
        fprintf(stdout, "     ");
      }
    vfprintf(stdout, format, v);
    va_end(v);
    fprintf(stdout, "\n");
  }


      /****************************************************************
       **                     LIBRARY FUNCTIONS                      **
       ****************************************************************/

static char Libraries_Included[10][20];
static int Lib_Index = 0;

void record_library(const char *libname)
  {
    strncpy(Libraries_Included[Lib_Index], libname, 20);
    Libraries_Included[Lib_Index][19] = '\0';
    ++Lib_Index;
  }


void show_avail_libs(void)
  {
    struct ffblk forge;
    int success;

    printf("    Available system libraries (modules):\n");

    success = findfirst("*.inc", &forge, 0);
    while (success == 0)
      {
        printf("      .\\%s\n", strlwr(forge.ff_name));
        success = findnext(&forge);
      }

    success = findfirst("lib\\*.inc", &forge, 0);
    while (success == 0)
      {
        printf("      lib\\%s\n", strlwr(forge.ff_name));
        success = findnext(&forge);
      }
  }


void dump_libs(void)
  {
    int i, c;

    if (Lib_Index)
      {
        assemble("PUSH  DS");
        assemble("POP   ES");
      }

    for (i=0; i < Lib_Index; i++)
      {
        FILE *fp;
        char buf[40];
        char *Lib = Libraries_Included[i];

        buf[0] = '\0';
        if (strchr(Lib, '\\') == 0)             /* figure out lib's name */
          strcpy(buf, "lib\\");
        strcat(buf, Lib);
        if (strchr(Lib, '.') == 0)
          strcat(buf, ".inc");

        fp = fopen(buf, "r");               /* Then dump the whole thing */

        if (!fp)
          {
            fprintf(stderr, "Library not found: %s\n", buf);
            continue;
          }

        printf(";;;; Library module: %s\n", buf);

        while ((c=getc(fp)) != EOF)
          putchar(c);
      }
  }


